/*

ISO 6969

*/


#include <sys/time.h>
#include <unistd.h>

#include "gaul.h"

#include "var.h"


#undef DEBUG

#define MAX_NUCLEO MAX_CHROMO



char* crackaddr(char* addr);


struct nucleo
{
  int n;
  char c;
  int last;
};


typedef struct nucleo chromo_t[MAX_NUCLEO];

long evaluation_count=0; 



boolean chromo_allocate(population *pop, entity *embryo)
{
int i;

  if (!pop) die("Null pointer to population structure passed.");
  if (!embryo) die("Null pointer to entity structure passed.");

  if (embryo->chromosome!=NULL)
    die("This entity already contains chromosomes.");

  embryo->chromosome = s_malloc(pop->num_chromosomes*sizeof(void *));

  for (i=0; i<pop->num_chromosomes; i++)
    embryo->chromosome[i]=s_malloc(pop->len_chromosomes*sizeof(struct nucleo));

  return TRUE;
}


void chromo_deallocate(population *pop, entity *corpse)
{
int i;

  if (!pop) die("Null pointer to population structure passed.");
  if (!corpse) die("Null pointer to entity structure passed.");

  if (corpse->chromosome==NULL)
    die("This entity already contains no chromosomes.");

  for (i=0; i<pop->num_chromosomes; i++)
    s_free(corpse->chromosome[i]);

  s_free(corpse->chromosome);

  corpse->chromosome=NULL;

  return;
}


void chromo_replicate( population *pop,
                                      entity *parent, entity *child,
                                      const int chromosomeid )
{

  if (!pop) die("Null pointer to population structure passed.");
  if (!parent || !child) die("Null pointer to entity structure passed.");
  if (!parent->chromosome || !child->chromosome) die("Entity has no chromsomes.");

  memcpy( child->chromosome[chromosomeid],
          parent->chromosome[chromosomeid],
          pop->len_chromosomes * sizeof(struct nucleo));

  return;
}



unsigned int chromo_to_bytes(population *pop, entity *joe,
                                     byte **bytes, unsigned int *max_bytes)
{
int           num_bytes;      /* Actual size of genes. */

  if (!pop) die("Null pointer to population structure passed.");
  if (!joe) die("Null pointer to entity structure passed.");

  if (*max_bytes!=0) die("Error: chromo_to_bytes max_bytes!=0");

  if (!joe->chromosome)
  {
    *bytes = (byte *)"\0";
    return 0;
  }

  num_bytes = pop->len_chromosomes * pop->num_chromosomes *
              sizeof(struct nucleo);

  *bytes = (byte *)joe->chromosome[0];

  return num_bytes;
}


void chromo_from_bytes(population *pop, entity *joe, byte *bytes)
{

  if (!pop) die("Null pointer to population structure passed.");
  if (!joe) die("Null pointer to entity structure passed.");

  if (!joe->chromosome) die("Entity has no chromsomes.");

  memcpy(joe->chromosome[0], bytes,
         pop->len_chromosomes * pop->num_chromosomes * sizeof(struct nucleo));

  return;
}



char* chromo_to_string(
                              const population *pop, const entity *joe,
                              char *text, size_t *textlen)
{
int i, j;
int n,nn;
int t;
char cc;
int last;


  if (!pop) die("Null pointer to population structure passed.");
  if (!joe) die("Null pointer to entity structure passed.");

  if ( (!text) || (*textlen < MAX_CHROMO ) )
  {
    *textlen = MAX_CHROMO;
    text = s_realloc(text,*textlen);
  }


  if (!joe->chromosome)
  {
    *text=0;
    return text;
  }

  t=0;
  for(i=0; i<pop->num_chromosomes; i++)
  {
    for(j=0; j<pop->len_chromosomes; j++)
    {
      last=((*((chromo_t*)(joe->chromosome[i])))[j]).last;
      if (last) break;
      nn=((*((chromo_t*)(joe->chromosome[i])))[j]).n;
      cc=((*((chromo_t*)(joe->chromosome[i])))[j]).c;
      if (t+nn >= MAX_CHROMO) break;
      for (n=0;n<nn;n++) text[t+n]=cc;
      t+=nn;
    }
  }     
  text[t]=0;

  return text;
}



int chromo_print(const population* pop,const entity* joe)
{
unsigned char* str;
int len;
int clen;
int c;


  str=NULL;
  len=0;
  str=chromo_to_string(pop,joe,str,&len);

  clen=chromo_len(pop,joe,0);

  for (c=0;c<len;c++)
  {
    if (str[c]<0x20 || str[c]>0x7f) printf("\\%02x",str[c]);
    else printf("%c",str[c]);
  }
  printf("\n");

//  printf("%i: %s\n",clen,str);

  s_free(str);

  return 0;
}

  

int chromo_len(population* pop,entity* joe,int cnum)
{
int i;
int last;

//  return pop->len_chromosomes;

  for(i=0; i<pop->len_chromosomes; i++)
  {
    last=((*((chromo_t*)(joe->chromosome[cnum])))[i]).last;
    if (last) break;
  }

  if (!last)
  {
    printf("Error in chromo_len: last not found\n");
  }

  return i;
}




int chromo_simplify(chromo_t chromo)
{
int c,d;
int zooo;
int size;

  if (NO_SIMPLIFY == 1) return 0;

  if ((chromo[0].last)) return 0;

  size=0;
  for (c=0;!(chromo[c+1].last);c++)
  {
    zooo=0;

    if (chromo[c].n<=0)
    {
      zooo=1;
      c--;
    }
    else
    {
      if (chromo[c].c == chromo[c+1].c)
      {
        (chromo[c].n)+=(chromo[c+1].n);
        zooo=1;
      }

      size+=(chromo[c].n);
    }
    
    if (zooo)
    {
      for (d=c+1;!(chromo[d+1].last);d++)
      {
        memcpy((void*)(&(chromo[d])),(void*)(&(chromo[d+1])),sizeof(chromo[d]));
      }
      chromo[d].last=1;
    }

    if (size>=MAX_CHROMO) chromo[c].last=1;
  }      

  return 0;
}



int chromo_insert_alelle(population* pop,entity* joe,int cnum,int pos,
			int nval,char cval)
{
int c,d;
int zooo;
int size;
struct nucleo* chromo;

  chromo=joe->chromosome[cnum];

//  printf("insert at pos %i n=%i c=%c\n",pos,nval,cval);

  for (d=0;!(chromo[d].last);d++)
    if (d>=MAX_NUCLEO) break;
  if (!(chromo[d].last))
  {
    printf("chromo_insert_alelle: cant find last.\n");
  }

//  printf("last at pos %i\n",d);
  if (d>=(MAX_NUCLEO-1)) return -1;


  chromo[d].last=0;
  chromo[d+1].last=1;

  for (;d>pos;d--)
  {
//    printf("copy %i to %i\n",d-1,d);
    memcpy((void*)(&(chromo[d])),(void*)(&(chromo[d-1])),sizeof(chromo[d]));
  }

  chromo[pos].n=nval;
  chromo[pos].c=cval;

  chromo_simplify(chromo);

  return 0;
}



char get_random_char()
{
char* clist="A();:.[]\"\\<>@";
int listlen;
char c;

  listlen=strlen(clist);

  if (REDUCE_CHARLIST==1) c=clist[random_int(listlen)];
  else do {c=random_int(127);} while (c<0x20);

  return c;
}



boolean score(population* pop,entity* entity)
{
char* ptr;
char* p;
char* str;
int len;
int bonus;


  str=NULL;
  len=0;
  str=chromo_to_string(pop,entity,str,&len);

  ptr=crackaddr(str);

  for (p=ptr+BUFSIZE;p>ptr;p--)
  {
    if (*p !=0 ) break;
//    if (*p =='A' ) break;
  }

  if (FITNESS_TYPE == 0) entity->fitness=p-ptr;
  if (FITNESS_TYPE == 1) entity->fitness=p-ptr-MAXNAME;

  if (SHORTER_BETTER ==1 )
  {
    bonus=(MAX_CHROMO - strlen(str));
    if (bonus > 2) bonus=2;
    entity->fitness+=bonus;
  }

  if (entity->fitness<0) entity->fitness=0;

  entity->fitness++;	// avoid bug when fitness==0

  evaluation_count++;

  return TRUE;
}



void mutate(population *pop, entity *father, entity *son)
{
int i;
int c;
int pos;
int pos2;
int len;
int dir;
int m,mm;


  if (!father || !son) die("Null pointer to entity structure passed");

  if (pop->len_chromosomes < 2) return;

/* Copy chromosomes of parent to offspring. */
  for (i=0; i<pop->num_chromosomes; i++)
  {
    memcpy(son->chromosome[i],father->chromosome[i],
			pop->len_chromosomes * sizeof(struct nucleo));
  }

  c=random_int(pop->num_chromosomes);

  len=chromo_len(pop,father,c);

// mm = number of mutation
  mm=random_int(MAX_MUTATION);
  if (mm<(MAX_MUTATION/2)) mm=1;

  for (m=0;m<mm;m++)
  switch (random_int(NB_MUTATION))
  {
    case 0: /* change alelle value */
//      printf("mutation 0 on chromosome %i\n",c);
      for (i=0;i<1;i++)
      {
        pos=random_int(len);
        ((*((chromo_t*)(son->chromosome[c])))[pos]).c=get_random_char();

        if (MUT_SIDE_EFFECT == 1)
        if (random_boolean() && (pos<(len-1)))
        {
          ((*((chromo_t*)(son->chromosome[c])))[pos+1]).c=get_random_char();
        }

      }
      break;

    case 1: /* insert new alelle */
//        printf("mutation 1 on chromosome %i\n",c);
        pos=random_int(len);
        chromo_insert_alelle(pop,son,c,pos,random_int(5),get_random_char());
      break;

    case 2: /* remove alelle */
//      printf("mutation 2 on chromosome %i\n",c);
      pos=random_int(len);
      ((*((chromo_t*)(son->chromosome[c])))[pos]).n=0;
      break;

    case 3: /* change alelle length */
//      printf("mutation 3 on chromosome %i\n",c);
      pos=random_int(len);
      ((*((chromo_t*)(son->chromosome[c])))[pos]).n+=(random_int(5)-3);
      break;

    case 4: /* copy several sequence from somewhere to elsewhere */
//      printf("mutation 4 on chromosome %i\n",c);

	dir=random_int(4)+1;
        if (dir>=len) break;
        pos=random_int(len-dir);
        pos2=random_int(len-dir);
	for (i=0;i<dir;i++)
        {
          memcpy( (void*)((struct nucleo*)(son->chromosome[c])+pos+i),
                  (void*)((struct nucleo*)(son->chromosome[c])+pos2+i),
                  sizeof(struct nucleo) );
        }

      break;

    case 5: /* add 1 to allele length and sub 1 to next allele */
//      printf("mutation 5 on chromosome %i\n",c);
      for (pos=0;pos < len-1;pos++)
      {
        dir=random_boolean()?-1:1;
        ((*((chromo_t*)(son->chromosome[c])))[pos]).n+=dir;
        ((*((chromo_t*)(son->chromosome[c])))[pos+1]).n-=dir;
      }
      break;


    default:
      printf("Alert! mutation under radioactivity !\n");
      break; 

  }

  chromo_simplify(son->chromosome[c]);

#ifdef DEBUG
  printf("mutate chomosome %i:\nfather:",c);
  chromo_print(pop,father);
  printf("** son:");
  chromo_print(pop,son);
  printf("\n");
#endif

  return;
}



static void crossover( population *pop,
                                 entity *father, entity *mother,
                                 entity *son, entity *daughter )
{
int location;
int i;
int len,lenf,lenm;

  if (!father || !mother || !son || !daughter)
    die("Null pointer to chromosome structure passed.");

  for (i=0; i<pop->num_chromosomes; i++)
  {
    lenm=chromo_len(pop,mother,i);
    lenf=chromo_len(pop,father,i);
    len=min(lenm,lenf);
    if (len < 2) break;

    location=random_int(len-1);

    memcpy((void*)(son->chromosome[i]),(void*)(mother->chromosome[i]),
				location*sizeof(struct nucleo));
    memcpy((void*)(daughter->chromosome[i]),(void*)(father->chromosome[i]),
				location*sizeof(struct nucleo));

    memcpy( (void*)((*((chromo_t*)(son->chromosome[i])))+location),
		(void*)((*((chromo_t*)(father->chromosome[i])))+location),
			(pop->len_chromosomes-location)*sizeof(struct nucleo));

    memcpy( (void*)((*((chromo_t*)(daughter->chromosome[i])))+location),
		(void*)((*((chromo_t*)(mother->chromosome[i])))+location),
			(pop->len_chromosomes-location)*sizeof(struct nucleo));
  }


#ifdef DEBUG
  printf("crossover location: %i\n",location);
  printf("crossover\nfather:");
  chromo_print(pop,father);
  printf("\nmother:");
  chromo_print(pop,mother);
  printf("\n** son:");
  chromo_print(pop,son);
  printf("\ndaughter:");
  chromo_print(pop,daughter);
  printf("\n");
#endif

  return;
}



entity* adapt(population *pop, entity *child)
{
entity* adult;

  adult = ga_entity_clone(pop,child);

  mutate(pop,child,adult);
  score(pop,adult);

  if (adult->fitness > child->fitness) return adult;

  ga_entity_copy_all_chromosomes(pop,adult,child);
  adult->fitness = child->fitness;

  return adult;
}



boolean seed(population *pop, entity *adam)
{
int c;
int p;
int len;
int n;

  if (!pop) die("Null pointer to population structure passed.");
  if (!adam) die("Null pointer to entity structure passed.");

  for (c=0; c<pop->num_chromosomes; c++)
  {
    len=0;
    for (p=0; p<(pop->len_chromosomes-1); p++)
    {
      n=1;
      if ((len+n) > MAX_CHROMO) break;
      ((*((chromo_t*)(adam->chromosome[c])))[p]).n=n;
      ((*((chromo_t*)(adam->chromosome[c])))[p]).c=get_random_char();
      ((*((chromo_t*)(adam->chromosome[c])))[p]).last=0;
      len+=n;
    }
    ((*((chromo_t*)(adam->chromosome[c])))[p]).last=1;

    chromo_simplify(adam->chromosome[c]);
  }

  return TRUE;
}



boolean generation_hook(int generation, population *pop)
{
static double total_best_fitnesses;
double average, stddev;
entity* bestone;
entity* lastone;
int p;
double bf;

  bestone=ga_get_entity_from_rank(pop,0);
  lastone=ga_get_entity_from_rank(pop,pop->size-1);

  total_best_fitnesses += bestone->fitness;

  if (generation%20 == 0)
  {
    printf("#############################################################\n");
    printf("Generation = %d\n", generation);

    if (FITNESS_TYPE==0) bf=bestone->fitness-MAXNAME-1;
    if (FITNESS_TYPE==1) bf=bestone->fitness-1;
    if (bf<0) bf=0;

    printf("Best fitness = %f\n",bf);
//    printf("Last fitness = %f\n", lastone->fitness);
/*
    chromo_print(pop,lastone);
    for (p=0;p<pop->size;p++)
    {
      chromo_print(pop,ga_get_entity_from_rank(pop,p));
    }
*/
    chromo_print(pop,bestone);

  }

/*
 * Stop if the population has converged.
 */

/*
  if (bestone->fitness == lastone->fitness)
  {
    printf("Solutions have converged!\n");
    return FALSE;
  }
*/

  return TRUE;  /* TRUE indicates that evolution should continue. */
}




population* pop_new(int size)
{
population* pop;
struct timeval tv;


  pop=ga_population_new(size,1,MAX_NUCLEO);
  if (!pop) return NULL;

  pop->generation_hook = generation_hook;
  pop->iteration_hook = NULL;

  pop->data_destructor = NULL;
  pop->data_ref_incrementor = NULL;

  pop->chromosome_constructor = chromo_allocate;
  pop->chromosome_destructor = chromo_deallocate;
  pop->chromosome_replicate = chromo_replicate;
  pop->chromosome_to_bytes = chromo_to_bytes;
  pop->chromosome_from_bytes = chromo_from_bytes;
  pop->chromosome_to_string = NULL;

  pop->evaluate = score;
  pop->seed = seed;
  pop->adapt = adapt;
  pop->select_one = ga_select_one_sus;
  pop->select_two = ga_select_two_sus;
  pop->mutate = mutate;
  pop->crossover = crossover;
  pop->replace = NULL;

  gettimeofday(&tv,NULL);
  random_seed(tv.tv_sec);
  ga_population_seed(pop);

  pop->scheme = GA_SCHEME_DARWIN;
  pop->elitism = GA_ELITISM_PARENTS_SURVIVE;

  pop->crossover_ratio = CROSSOVER_RATIO;
  pop->mutation_ratio = MUTATION_RATIO;
  pop->migration_ratio = MIGRATION_RATIO;

  return pop;
}



int main(int argc,char** argv)
{
int i;
population* poplist[100];
char* str;
int len;
entity* bestone;
int pop_num;

  pop_num=POP_NUM;

  for (i=0;i<pop_num;i++)
  {
    poplist[i]=pop_new(POP_SIZE);
    if (!poplist[i]) return -1;
    ga_population_set_scheme(poplist[i],SCHEME);
  }


  if (pop_num==1)  ga_evolution(poplist[0],MAX_GENERATION);
  else  ga_evolution_archipelago(pop_num,poplist,MAX_GENERATION);


  for (i=0;i<pop_num;i++)
  {
    bestone=ga_get_entity_from_rank(poplist[i],0);
    str=NULL;
    len=0;
    str=chromo_to_string(poplist[i],bestone,str,&len);
    printf( "The final solution on island %i is:\n%s\n",i,str);
    printf( "With score = %f\n",ga_entity_get_fitness(bestone));

    ga_extinction(poplist[i]);

    s_free(str);
  }

  return 0;
}

